/*
 *	AutoPan.c --- a module for fvwm
 *		pan among desktops automatically ---
 *
 * Copyright (c) 2002-2003
 * 	Koh-ichi Ito All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the developer nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include <ctype.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#ifdef FVWM2
#include "Module.h"	/* borrow from fvwm source for SendInfo() */
#else
#include "fvwmlib.h"	/* borrow from fvwm source for SendInfo() */
#endif
#define ROOT_WINDOW	( (unsigned long)0)	/* may be buggy */
#define BUFLEN	160
#define KEYLEN	40
#define INTERVAL	5
#define HORIZONTAL	0
#define VERTICAL	1

char	*MyName;
static char	Id[] = "$Id: AutoPan.c,v 1.3.1.1 2003/07/09 14:12:47 kohi Exp $";
char	*basename();
void	scroll();
int	parse_conf();
int	parse_arg();
void	exithandler();
void	usage();

struct param {
	int	interval,
		direction,
		size_x,
		size_y;
};

struct {
	int	fd;
	char	name[MAXPATHLEN];
} CtlFifo;

main(argc, argv)

	int	argc;
	char	**argv;
{
	int	fd[2];
	int	i, j;
	struct param	param;
	struct stat	st;

	signal(SIGPIPE, exithandler);
	signal(SIGTERM, exithandler);
	MyName = basename(argv[0] );
	if (argc < 4) {
		usage();
	}
	fd[0] = atoi(argv[1] );
	fd[1] = atoi(argv[2] );
	param.interval = INTERVAL;
	param.direction = HORIZONTAL;
	param.size_x = 3;	/* default value of */
	param.size_y = 3;	/* DeskTopSize is 3x3 */
	if (!parse_conf(argv[3], &param) ) {
		exit(1);
	}
	if ( (param.size_x <= 0) || (param.size_y <= 0) ) {
		fprintf(stderr, "%s: fail to parse DeskTopSize.\n", MyName);
		exit(1);
	} else if ( (param.size_x == 1) && (param.size_y == 1) ) {
		fprintf(stderr,
			"%s: DeskTopSize is 1x1, no mean for panning.\n",
			MyName);
		exit(1);
	}
	CtlFifo.name[0] = '\0';
	CtlFifo.fd = -1;
	if (6 < argc) {
		if (!parse_arg(argv[6], &param) ) {
			exit(1);
		}
	}
	if (CtlFifo.name[0] ) {
		if (stat(CtlFifo.name, &st) == -1) {
			if (mkfifo(CtlFifo.name, 0600) < 0) {
				warn("mkfifo(%s)", CtlFifo.name);
				CtlFifo.name[0] = '\0';
				CtlFifo.fd = -1;
			}
		} else {
			if (!(st.st_mode & S_IFIFO) ) {
				fprintf(stderr,
					"%s: %s exist and isn't a FIFO.\n",
					MyName, CtlFifo.name);
				CtlFifo.name[0] = '\0';
				CtlFifo.fd = -1;
			}
		}
	}
	if (CtlFifo.name[0] ) {
		/* CtlFifo.name[0] may be rewritten in above closure.*/
		if ( (CtlFifo.fd = open(CtlFifo.name, O_RDWR) ) < 0) {
			/* O_RDONLY is right, but it makes open()
			   to be blocked.*/
			warn("open(%s)", CtlFifo.name);
			CtlFifo.fd = -1;
			CtlFifo.name[0] = '\0';
		}
	}
	if (param.direction == HORIZONTAL) {
		for (;;) {
			for (j = 0; j < param.size_y; j++) {
				for (i = 0; i < param.size_x; i++) {
					scroll(i, j, param.interval, fd);
				}
			}
		}
	} else {
		for (;;) {
			for (i = 0; i < param.size_x; i++) {
				for (j = 0; j < param.size_y; j++) {
					scroll(i, j, param.interval, fd);
				}
			}
		}
	}
	exithandler();
}

void scroll(i, j, interval, fd)

	int	i, j,
		interval,
		fd[];
{
	char	s[BUFLEN],
		s2[BUFLEN];
	ssize_t	n;
	struct timeval	timeout;
	fd_set	ifd;

	snprintf(s, BUFLEN, "GotoPage %d %d", i, j);
	SendInfo(fd, s, ROOT_WINDOW);
	FD_ZERO(&ifd);
	if (0 < CtlFifo.fd) {
		FD_SET(CtlFifo.fd, &ifd);
	}
	timeout.tv_sec = (long)interval;
	timeout.tv_usec = (long)0;
	select(CtlFifo.fd + 1, &ifd, NULL, NULL, &timeout);
	if ( (0 < CtlFifo.fd)
	  && (FD_ISSET(CtlFifo.fd, &ifd) ) ) {
		if (0 < (n = read(CtlFifo.fd, s2, BUFLEN) ) ) {
			s2[n] = '\0';
			if (valid_format(s2) ) {
				snprintf(s, BUFLEN, "GotoPage %s", s2);
				SendInfo(fd, s, ROOT_WINDOW);
				sleep(interval);
			}
		}
	}
}

int parse_conf(filename, param)

	char	*filename;
	struct param	*param;
{
	char	s[BUFLEN],
		*key, *val, *p;
	int	i;
	FILE	*fp;

	if (!(fp = fopen(filename, "r") ) ) {
		snprintf(s, sizeof(s), "%s: paramater file \"%s\":",
			MyName, filename);
		perror(s);
		return 0;
	}
	key = s;	/* always same */
	while (0 < fgets(s, BUFLEN, fp) ) {
		/* s: KeyWord   Value */
		s[BUFLEN - 1] = '\0';	/* insurance */
		p = s;
		while (!isspace(*p) && (p < s + BUFLEN) && *p) {
			p++;
		}
		/*   s: KeyWord   Value */
		/*   p:        ^        */
		/* key: ^               */
		*p++ = '\0';
		/*   s: KeyWord\0 Value */
		/*   p:          ^      */
		/* key: ^               */
		/* now key points the key part of a line from .fvwmrc */
		while (isspace(*p) && (p < s + BUFLEN) ) {
			p++;
		}
		/*   s: KeyWord\0 Value */
		/*   p:           ^     */
		/* key: ^               */
		if (s + BUFLEN <= p) {
			fprintf(stderr,
				"%s: overrun occur while parsing entry \"%s\".\n",
				MyName, key);
			continue;
		}
		val = p;
		/*   s: KeyWord\0 Value */
		/*   p:           ^     */
		/* key: ^               */
		/* val:           ^     */
		if (strcmp(key, "DeskTopSize") == 0) {
			/* p = val; yes, always so */
			while ( (*p != 'x') && (p < s + BUFLEN) ) {
				p++;
			}
			*p++ = '\0';
			param->size_x = atoi(val);
			param->size_y = atoi(p);	/* believe insurance */
		}
	}
	fclose(fp);
	return 1;
}

int parse_arg(arg, param)

	char	*arg;
	struct param	*param;
{
	char	*p, *q,	*eq;
	int	eoln;

	eoln = 0;
	p = arg;
	while (!eoln) {
		q = p;
		while (*q && !isspace(*q) ) {
			q++;
		}
		if (!*q) {
			eoln = 1;
		} else {
			*q = '\0';
		}
		if (isdigit(*p) ) {
			param->interval = atoi(p);
		} else if (strcmp(p, "horizontal") == 0) {
			param->direction = HORIZONTAL;
		} else if (strcmp(p, "vertical") == 0) {
			param->direction = VERTICAL;
		} else if ( (eq = index(p, '=') )
		       && (strncmp(p, "ctl", strlen("ctl") ) == 0) ) {
			while (*++eq && isspace(*eq) )
				;
			strncpy(CtlFifo.name, eq, MAXPATHLEN);
		} else {
			fprintf(stderr, "%s: unknown argument %s.\n",
				MyName, p);
			/* return 0;	to make it an error */
		}
		if (!eoln) {
			p = q + 1;
			while (isspace(*p) ) {
				p++;
			}
			if (!*p) {
				eoln = 1;
			}
		}
	}
	return 1;
}

int valid_format(s)

	char	*s;
{
	while (*s && isspace(*s) ) {
		s++;
	}
	if (!isdigit(*s) ) {
		return 0;
	}
	while (*s && isdigit(*s) ) {
		s++;
	}
	if (!isspace(*s) ) {
		return 0;
	}
	while (*s && isspace(*s) ) {
		s++;
	}
	if (!isdigit(*s) ) {
		return 0;
	}
	while (*s && isdigit(*s) ) {
		s++;
	}
	if (!*s) {
		return 1;
	}
	while (*s) {
		if (!isspace(*s) ) {
			return 0;
		}
		s++;
	}
	return 1;
}

char *basename(path)

	char	*path;
{
	char	*p;

	p = path + strlen(path) - 1;
	while ( (*p != '/') && (path < p) ) {
		p--;
	}
	if (*p == '/') {
		return p + 1;
	} else {
		return path;
	}
}

void exithandler()
{
	if (CtlFifo.name[0] ) {
		close(CtlFifo.fd);
		unlink(CtlFifo.name);
	}
	exit(0);
}

void usage()
{
	fprintf(stderr,
		"Usage: %s must be invoked via fvwm as a Module.\n",
		MyName);
	exit(1);
}
